'use strict';

/**
 * @abstract
 * Create a new Launcher instance.
 * @param {Game} game - The game in which the launcher is placed
 * @param {Number} x - The launcher's starting x-coordinate
 * @param {Number} y - The launcher's starting y-coordinate
 */
function Launcher(game, x, y) {
	this._game = game;
	
	this._body = new Sprite(TextureCache[this.BODY_SPRITE]);
	
	/** @private {Number} The launcher's x-coordinate in the world, not including screen scroll */
	this._baseX;
	
	this.x = x;
	this.y = y;
	
	if (this.LAUNCH_SOUND) {
		this._launchSound = loader.resources[this.LAUNCH_SOUND].data;
	}
	
	this._currentFrame = 0;
	
	this._state = this.STATES.IDLE;
}

// Define constants.
/** @constant {String} The path from the game root to the sound effect for launching the projectile */
Launcher.prototype.LAUNCH_SOUND = '';
/** @constant {String} The path from the game root to the sprite for the launcher body */
Launcher.prototype.BODY_SPRITE = 'images/game/launcher_body.png';
/** @constant {Number} The duration of the launch animation, in frames */
Launcher.prototype.LAUNCH_DURATION = 20;

Launcher.prototype.STATES = {
	IDLE: 0,
	LAUNCH: 1
};

/**
 * Queue assets to be preloaded that might be needed by any Launcher instance.
 */
Launcher.queueAssets = function () {
	loader.add(this.prototype.BODY_SPRITE);
	if (this.prototype.ARM_SPRITE) {
		loader.add(this.prototype.ARM_SPRITE);
	}
	if (this.prototype.LAUNCH_SOUND) {
		loader.add(this.prototype.LAUNCH_SOUND);
	}
};

Object.defineProperties(Launcher.prototype, {
	x: {
		get: function () {
			return this._baseX;
		},
		set: function (value) {
			this._body.x = value - this._game.scrollX;
			if (this._arm) {
				this._arm.x = value + this.ARM_X - this._game.scrollX;
			}
			this._baseX = value;
		}
	},
	y: {
		get: function () {
			return this._body.y;
		},
		set: function (value) {
			this._body.y = value;
			if (this._arm) {
				this._arm.y = value + this.ARM_Y;
			}
		}
	}
});

/**
 * Trigger the launch animation and sound.
 */
Launcher.prototype.launch = function () {
	if (this._launchSound && !this._game.muted) {
		this._launchSound.currentTime = 0;
		this._launchSound.play();
	}
	this._state = this.STATES.LAUNCH;
};

/**
 * Update the launcher for the next frame.
 * @param {Number} deltaTime - The time since the last frame, in milliseconds
 */
Launcher.prototype.update = function (deltaTime) {
	// Update x in case the screen has scrolled.
	this.x = this.x;
	
	if (this._state === this.STATES.LAUNCH) {
		this._currentFrame++;
		if (this._currentFrame >= this.LAUNCH_DURATION) {
			this._currentFrame = 0;
			this._state = this.STATES.IDLE;
		}
	}
};
