'use strict';

var Utils = {
	
	/** {Boolean} Whether the game is running on Mobile Safari */
	isMobileSafari: !!navigator.userAgent.match(/ip(ad|hone|od)/i),
	
	/**
	 * Calculate the value for a given `x` from 0 to `xMax` where `y(x = 0) = 0`, `y(x = 0.5 * xMax) = 1`, and `y(x = xMax) = 0`.
	 * @param {Number} x - The current x-value
	 * @param {Number} xMax - The maximum x-value
	 * @returns {Number} A value from 0 to 1
	 */
	absFunc: function (x, xMax) {
		return (-Math.abs((2 / xMax) * x - 1) + 1);
	},
	
	/**
	 * Constrain a value between a minimum and maximum.
	 * @param {Number} value - The value to constrain
	 * @param {Number} min - The minimum value to allow
	 * @param {Number} max - The maximum value to allow
	 * @returns {Number} `value` or the closest number between `min` and `max`
	 */
	constrainValue: function (value, min, max) {
		return Math.max(min, Math.min(max, value));
	},
	
	/**
	 * Load a file.
	 * @param {String} path - The path to the file
	 * @returns {Promise} Resolves with the contents of the file or an error message
	 */
	fetch: function (path) {
		return new Promise(function (resolve, reject) {
			var xhr = new XMLHttpRequest();
			xhr.onreadystatechange = function () {
				if (xhr.readyState === 4) {
					if (xhr.status === 200) {
						resolve(xhr.responseText);
					} else {
						reject('Error ' + xhr.status + ' while attempting to load ' + path);
					}
				}
			};
			xhr.open('GET', path, true);
			xhr.send();
		});
	},
	
	/**
	 * Get the x-coordinate of a click within an element.
	 * @param {Number} pageX - The x-coordinate relative to the page
	 * @param {HTMLElement} element - The element that was clicked
	 */
	getCanvasX: function (pageX, element) {
		return pageX - element.offsetLeft;
	},
	
	/**
	 * Get the y-coordinate of a click within an element.
	 * @param {Number} pageY - The y-coordinate relative to the page
	 * @param {HTMLElement} element - The element that was clicked
	 */
	getCanvasY: function (pageY, element) {
		return pageY - element.offsetTop;
	},
	
	/**
	 * Calculate the value for a given `x` from 0 to `xMax` where `y(x = 0) = yMin`, `y(x = 0.5 * xMax) = yMax`, and `y(x = xMax) = yMin`.
	 * @param {Number} x
	 * @param {Number} xMax
	 * @param {Number} yMin
	 * @param {Number} yMax
	 * @returns {Number} A value from `yMin` to `yMax`
	 */
	mappedAbsFunc: function (x, xMax, yMin, yMax) {
		return this.mapToRange(this.absFunc(x, xMax), yMin, yMax);
	},
	
	/**
	 * Map a value from 0 to 1 to a range from `min` to `max`.
	 * @param {Number} value
	 * @param {Number} min
	 * @param {Number} max
	 * @returns {Number} A value from `min` to `max`
	 */
	mapToRange: function (value, min, max) {
		return (min + value * (max - min));
	},
	
	/**
	 * Check whether a circle is touching a rectangle.
	 * @param {Number} cx - The circle's x-coordinate
	 * @param {Number} cy - The circle's y-coordinate
	 * @param {Number} cr - The circle's radius
	 * @param {Number} rx - THe rectangle's x-coordinate
	 * @param {Number} ry - THe rectangle's y-coordinate
	 * @param {Number} rw - THe rectangle's width
	 * @param {Number} rh - THe rectangle's height
	 * @returns {Boolean} - Whether the circle is touching the rectangle
	 */
	isCircCollidingRect: function (cx, cy, cr, rx, ry, rw, rh) {
		Array.from(arguments).forEach(function (argument) {
			if (typeof(argument) !== 'number' || isNaN(argument)) {
				throw new TypeError('All arguments to Utils.isCircCollidingRect must be valid Numbers.');
			}
		});
		
		// Calculate the distance between the two shapes.
		var distX = Math.abs(cx - rx - rw / 2);
		var distY = Math.abs(cy - ry - rh / 2);
		
		// If the distance is greater than half of each shape they are too far.
		if (distX > (rw / 2 + cr)) {
			return false;
		}
		if (distY > (rh / 2 + cr)) {
			return false;
		}
		
		// If the distance is less than half then they are colliding.
		if (distX <= (rw / 2)) {
			return true;
		} 
		if (distY <= (rh / 2)) {
			return true;
		} 
		
		// Check the last case for the corners.
		// Pythagorean formula for the lines from center to corners
		var dx = distX - rw / 2;
		var dy = distY - rh / 2;
		return (dx * dy + dy * dy <= (cr *  cr));
	},
	
	/**
	 * Check whether two rectangles are colliding.
	 * @param {Number} x1
	 * @param {Number} y1
	 * @param {Number} w1
	 * @param {Number} h1
	 * @param {Number} x2
	 * @param {Number} y2
	 * @param {Number} w2
	 * @param {Number} h2
	 * @returns {Boolean}
	 */
	isRectCollidingRect: function (x1, y1, w1, h1, x2, y2, w2, h2) {
		Array.from(arguments).forEach(function (argument) {
			if (typeof(argument) !== 'number' || isNaN(argument)) {
				throw new TypeError('All arguments to Utils.isRectCollidingRect must be Numbers.');
			}
		});
		
		return (x1 < x2 + w2 &&
			x1 + w1 > x2 &&
			y1 < y2 + h2 &&
			h1 + y1 > y2) 
	},
	
	/**
	 * Pick a random element from an array.
	 * @param {Array} array - The array to pick from
	 * @returns {Object} The picked array element
	 */
	pickRandom: function (array) {
		var index = Utils.randInt(0, array.length);
		return array[index];
	},
	
	/**
	 * Get a random integer greater than or equal to `min` and less than `max`.
	 * @param {Number} min
	 * @param {Number} max
	 * @returns {Number}
	 */
	randInt: function(min, max) {
		Array.from(arguments).forEach(function (argument) {
			if (typeof(argument) !== 'number' || isNaN(argument)) {
				throw new TypeError('All arguments to Utils.randInt must be Numbers.');
			}
		});
		
		return Math.floor(Math.random() * (max - min) + min);
	}
};
