'use strict';

/**
 * Create a new Enemy instance.
 * @param {Game} game - The game in which the enemy is to be placed
 * @param {Number} x - The enemy's starting x-coordinate
 * @param {Number} level - The game level of the enemy
 * @param {Number} variant - Which enemy type to use
 */
function Enemy(game, x, level, variant) {
	this._game = game;
	
	this._config = this.TYPES[variant];
	
	/** @private {Number} The game level of the enemy */
	this._level = 0;
	
	/** @private {Number} How much health before the enemy dies */
	this._health = 1;
	
	/** @private {String} This enemy instance's movement pattern: linear, sinusoidal, or random */
	this._movement = 'linear';
	
	// Copy the Pixi texture for each enemy type in case this enemy is reused later.
	this._spriteSheets = [];
	this.TYPES.forEach(function (config) {
		if (Utils.isMobileSafari && config.lowResSpriteSheet) {
			this._spriteSheets.push(TextureCache[config.lowResSpriteSheet].clone());
		} else {
			this._spriteSheets.push(TextureCache[config.spriteSheet].clone());
		}
	}, this);
	// Apply the texture for the requested variant.
	this._sprite = new Sprite(this._spriteSheets[(variant || 0)]);
	this._textureFrame = new Rectangle(0, 0, 1, 1);
	
	/** {Rectangle} The area in which the enemy can attack and is vulnerable to attacks */
	this.hitbox = new Rectangle(0, 0, 1, 1);
	
	/** @private {Number} The enemy's x-coordinate in the world, not including screen scroll */
	this._baseX = x;
	/** @private {Number} The enemy's starting y-coordinate, not including odd movement patterns */
	this._baseY;
	
	/** @private {Number} The current frame in the enemy's current animation */
	this._currentFrame = 0;
	
	this._state = this.STATES.ALIVE;
	
	// Bring the new enemy to life.
	this.revive(level, variant);
}

// Define constants.
/**
 * @constant {Array<Object<String,Object>>} Configurations for each enemy type:
 *     {String} spriteSheet - The path from the game root to the sprites for this enemy type
 *     {Number} spriteWidth - The width of each frame, in pixels
 *     {Number} spriteHeight - The height of each frame, in pixels
 *     {Number} bottomOffset - How far from the bottom of the game screen to spawn, in pixels
 *     {Number} hitboxX - The x-offset of the enemy's hitbox from the left side of the sprite
 *     {Number} hitboxY - The y-offset of the enemy's hitbox from the top of the sprite
 *     {Number} hitboxWidth - The width of the enemy's hitbox
 *     {Number} hitboxHeight - The height of the enemy's hitbox
 *     {Number} walkAnimLength - The length of the enemy's walk animations, in frames
 *     {Number} deathAnimLength - The length of the enemy's death animations, in frames
 *     {Number} frameRateMultiplier - The fraction of the frame rate the enemy animates at, from 0 to 1
 *     {Number} maxHealth - The maximum amount of health the enemy has sprites for
 *     {Array<String>} movement - The enemy's possible movement patterns: linear, sinusoidal, or random
 */
Enemy.prototype.TYPES = [
	{
		spriteSheet: 'images/game/enemy.png',
		spriteWidth: 142,
		spriteHeight: 196,
		bottomOffset: 8,
		hitboxX: 0,
		hitboxY: 0,
		hitboxWidth: 142,
		hitboxHeight: 196,
		walkAnimLength: 24,
		deathAnimLength: 8,
		frameRateMultiplier: 1,
		maxHealth: 1,
		movement: ['linear']
	}
];
/** @constant {String} The speed at which the enemy moves toward the player, in pixels per frame */
Enemy.prototype.SPEED = 1;
/** @constant {Number} The amount the enemy speed increases per game level */
Enemy.prototype.LEVEL_SPEED_INCREASE = 0.2;

Enemy.prototype.STATES = {
	ALIVE: 0,
	DYING: 1,
	DEAD: 2
};

/**
 * Queue assets to be preloaded that might be needed by any Enemy instance.
 */
Enemy.queueAssets = function () {
	this.prototype.TYPES.forEach(function (config) {
		if (Utils.isMobileSafari && config.lowResSpriteSheet) {
			loader.add(config.lowResSpriteSheet);
		} else {
			loader.add(config.spriteSheet);
		}
	});
};

Object.defineProperties(Enemy.prototype, {
	x: {
		get: function () {
			return this._baseX;
		},
		set: function (value) {
			this._baseX = value;
			this._sprite.x = value - this._game.scrollX;
			this.hitbox.x = value + this._config.hitboxX;
		}
	},
	y: {
		get: function () {
			return this._sprite.y;
		},
		set: function (value) {
			this._sprite.y = value;
			this.hitbox.y = value + this._config.hitboxY;
		}
	},
	width: {
		get: function () {
			return this._sprite.width;
		}
	},
	height: {
		get: function () {
			return this._sprite.height;
		}
	},
	isDead: {
		get: function () {
			return (this._state === this.STATES.DEAD);
		}
	},
	isDyingOrDead: {
		get: function () {
			return this._health === 0 && (this._state === this.STATES.DYING || this._state === this.STATES.DEAD);
		}
	}
});

/**
 * Kill the enemy and set its death animation to play.
 * @param {Boolean} skipAnimation - Whether to skip the death animation
 */
Enemy.prototype.kill = function (skipAnimation) {
	this._health = Math.max(0, this._health - 1);
	if (skipAnimation || this._config.deathAnimLength === 0) {
		this._state = this.STATES.DEAD;
		this._game.app.stage.removeChild(this._sprite);
	} else {
		this._state = this.STATES.DYING;
		this._currentFrame = 0;
		this._textureFrame.y = ((2 * this._health) + 1) * this._config.spriteHeight;
	}
};

/**
 * Restore the enemy to its default state.
 * @param {Number} level - The game level of the enemy
 * @param {Number} variant - Which sprite sheet to use
 */
Enemy.prototype.revive = function (level, variant) {
	this._config = this.TYPES[variant];
	
	this._level = level || 0;
	
	this._health = 1;
	if (Math.random() < (this._level / 10)) {
		// Increase the chance of 2-health enemies appearing until level 10,
		// at which point all enemies that can have 2 health will have 2 health.
		this._health = Math.min(2, this._config.maxHealth);
	}
	
	this._movement = this._config.movement[0];
	if (Math.random() < (this._level / 10)) {
		// Increase the chance of different enemy movement until level 10,
		// at which point all enemies can have different movement.
		this._movement = Utils.pickRandom(this._config.movement);
	}
	
	this.x = this._baseX;
	this.y =
		this._baseY = game.SCREEN_HEIGHT - this._config.spriteHeight - this._config.bottomOffset;
	
	this.hitbox.width = this._config.hitboxWidth;
	this.hitbox.height = this._config.hitboxHeight;
	
	// Start on a random frame so enemies do not walk in sync.
	if (Utils.isMobileSafari && this._config.lowResSpriteSheet) {
		this._currentFrame = Utils.randInt(0, this._config.lowResWalkAnimLength);
	} else {
		this._currentFrame = Utils.randInt(0, this._config.walkAnimLength);
	}
	
	// Apply the texture for the requested variant.
	this._sprite.texture = this._spriteSheets[(variant || 0)];
	this._textureFrame.x = Math.floor(this._currentFrame) * this._config.spriteWidth;
	this._textureFrame.y = (2 * (this._health - 1) * this._config.spriteHeight);
	this._textureFrame.width = this._config.spriteWidth;
	this._textureFrame.height = this._config.spriteHeight;
	this._sprite.texture.frame = this._textureFrame;
	
	this._state = this.STATES.ALIVE;
	this._game.app.stage.addChild(this._sprite);
};

/**
 * Update the enemy for the next frame.
 * @param {Number} deltaTime - The time since the last frame, in milliseconds
 */
Enemy.prototype.update = function (deltaTime) {
	// Do not update dead enemies.
	if (this.isDead) {
		return;
	}
	
	var frameRateMultiplier =
			(Utils.isMobileSafari && this._config.lowResSpriteSheet) ?
				this._config.lowResFrameRateMultiplier : this._config.frameRateMultiplier;
	this._currentFrame += frameRateMultiplier;
	if (this._state === this.STATES.ALIVE) {
		// Move across the screen.
		this.x -= this.SPEED + (this.LEVEL_SPEED_INCREASE * this._level);// * deltaTime;
		
		switch (this._movement) {
			case 'sinusoidal':
				this.y = 24 * Math.sin(0.05 * this.x) + this._baseY;
				break;
			case 'random':
				this.y = Utils.constrainValue(
					this.y + Utils.randInt(-1, 2),
					this._baseY - this._config.spriteHeight,
					this._baseY + this._config.spriteHeight);
				break;
		}
		
		// Advance the walk cycle.
		var walkAnimLength =
			(Utils.isMobileSafari && this._config.lowResSpriteSheet) ?
				this._config.lowResWalkAnimLength : this._config.walkAnimLength;
		if (Math.floor(this._currentFrame) >= walkAnimLength) {
			this._currentFrame = 0;
		}
	} else if (this._state === this.STATES.DYING) {
		// Update x in case the screen has scrolled.
		this.x = this.x;
		// At the end of the death animation, stay dead.
		var deathAnimLength =
			(Utils.isMobileSafari && this._config.lowResSpriteSheet) ?
				this._config.lowResDeathAnimLength : this._config.deathAnimLength;
		if (Math.floor(this._currentFrame) >= deathAnimLength) {
			if (this._health > 0) {
				// If it is not completely dead, let it resume moving.
				this._currentFrame = 0;
				this._textureFrame.y = (2 * (this._health - 1) * this._config.spriteHeight);
				this._state = this.STATES.ALIVE;
			} else {
				// If it is dead, remove it.
				this.kill(true);
				return;
			}
		}
	}
	
	// Get the new frame from the sprite sheet.
	this._textureFrame.x = Math.floor(this._currentFrame) * this._config.spriteWidth;
	this._sprite.texture.frame = this._textureFrame;
};
