'use strict';

/**
 * Create a new Projectile instance.
 * @param {Game} game - The game in which the projectile is going to exist
 * @param {Number} x - The projectile's starting x-coordinate
 * @param {Number} y - The projectile's starting y-coordinate
 */
function Projectile(game, x, y) {
	this._game = game;
	
	this._sprite = new Sprite(TextureCache[this.SPRITE]);
	this._sprite.anchor.x = 0.5;
	this._sprite.anchor.y = 0.5;
	this._spriteFrame = new Rectangle(0, 0, this.FRAME_WIDTH, this.FRAME_HEIGHT);
	this._sprite.texture.frame = this._spriteFrame;
	if (this.HIDDEN_UNTIL_LAUNCH) {
		this._sprite.visible = false;
	}
	
	/** @private {Number} The projectile's x-coordinate in the world, not including screen scroll */
	this._baseX;
	
	this._startX =
		this.x = x;
	this._startY =
		this.y = y;
	
	if (this.HIT_SOUND) {
		this._hitSound = loader.resources[this.HIT_SOUND].data;
	}
	
	this.velocity = new Vector2D(0, 0);
	
	/** @private {Number} How long into the launcher animation to wait to launch the projectile, in frames */
	this._launchDelayTimer = 0;
	
	this._state = this.STATES.IDLE;
	
	game.app.stage.addChild(this._sprite);
}

// Define constants.
/** @constant {String} The path from the game root to the sound effect for the projectile hitting a target or the ground */
Projectile.prototype.HIT_SOUND = '';
/** @constant {String} The path from the game root to the sprite for the projectile */
Projectile.prototype.SPRITE = 'images/game/projectile.png';
/** @constant {String} The width of the projectile, in pixels */
Projectile.prototype.FRAME_WIDTH = 60;
/** @constant {String} The height of the projectile, in pixels */
Projectile.prototype.FRAME_HEIGHT = 60;
/** @constant {String} The number of projectile variants in the sprite sheet */
Projectile.prototype.VARIANTS = 1;
/** @constant {Number} How far through the launch animation to launch the projectile, from 0 to 1 */
Projectile.prototype.LAUNCH_DELAY = 0;
/** @constant {Boolean} Whether the projectile is hidden while in the launcher */
Projectile.prototype.HIDDEN_UNTIL_LAUNCH = false;

Projectile.prototype.STATES = {
	IDLE: 0,
	LAUNCHED: 1
};

/**
 * Queue assets to be preloaded that might be needed by any Projectile instance.
 */
Projectile.queueAssets = function () {
	loader.add(this.prototype.SPRITE);
	if (this.prototype.HIT_SOUND) {
		loader.add(this.prototype.HIT_SOUND);
	}
};


Object.defineProperties(Projectile.prototype, {
	x: {
		get: function () {
			return this._baseX;
		},
		set: function (value) {
			this._sprite.x = value - this._game.scrollX;
			this._baseX = value;
		}
	},
	y: {
		get: function () {
			return this._sprite.y;
		},
		set: function (value) {
			this._sprite.y = value;
		}
	},
	radius: {
		get: function () {
			return this._sprite.width / 2;
		}
	}
});

/**
 * Launch the projectile.
 * @param {Vector2D} launchVelocity - The velocity of the projectile at launch
 */
Projectile.prototype.launch = function (launchVelocity) {
	this.velocity = launchVelocity;
	this._launchDelayTimer = this.LAUNCH_DELAY * Launcher.prototype.LAUNCH_DURATION;
	this._state = this.STATES.LAUNCHED;
};

/**
 * Reset the projectile to its starting position and visibility.
 */
Projectile.prototype.reset = function () {
	// Play the hit sound, if any.
	if (this._hitSound && !this._game.muted) {
		this._hitSound.currentTime = 0;
		this._hitSound.play();
	}
	
	// Change to a different variant.
	this._spriteFrame.x = Utils.randInt(0, this.VARIANTS) * this.FRAME_WIDTH;
	this._sprite.texture.frame = this._spriteFrame;
	
	// Reset the position.
	this.x = this._startX;
	this.y = this._startY;
	
	// Reset the visibility.
	if (this.HIDDEN_UNTIL_LAUNCH) {
		this._sprite.visible = false;
	}
	
	// Set it as idle again.
	this._state = this.STATES.IDLE;
};

/**
 * Update the projectile for the next frame.
 * @param {Number} deltaTime - The time since the last frame, in milliseconds
 */
Projectile.prototype.update = function (deltaTime) {
	// Update x in case the screen has scrolled.
	this.x = this.x;
	
	// Do not update the projectile if it is idle.
	if (this._state === this.STATES.IDLE) {
		return;
	}
	
	// Delay the launch to partway through the launcher animation.
	if (this._launchDelayTimer > 0) {
		this._launchDelayTimer--;
		return;
	}
	
	// If the projectile has been launched, make sure it is visible.
	this._sprite.visible = true;
	
	this.velocity.incrementBy(this._game.gravity);
	this.x += this.velocity.x;
	this.y += this.velocity.y;
};
